<html>
  <head>
    <title>WebDAV Principal - {{ principal.get_displayname() }}</title>
  </head>
  <body>
    <h1>{{ principal.get_displayname() }} </h1>

    <p>This is a user principal. CalDAV/CardDAV clients that support
    autodiscovery can use the URL for this page for discovery.
    </p>

    <h2>Subcollections</h2>

    <ul>
    {% for name, resource in principal.subcollections() %}
       <li><a href="{{ urljoin(self_url+'/', name+'/') }}">{{ name }}</a></li>
    {% endfor %}
    </ul>

    <p>For more information about Xandikos, see <a
    href="https://www.xandikos.org/">https://www.xandikos.org/</a>
    or <a href="https://github.com/jelmer/xandikos">https://github.com/jelmer/xandikos</a>.
    </p>
  </body>
</html>
